#
# Copyright 3D Control Systems, Inc. All Rights Reserved 2017-2019. Built in San Francisco.
#
# This software is distributed under commercial non-GPL license for personal, educational,
# corporate or any other use. The software as a whole or any parts of that are prohibited
# for distribution and/or use without obtaining license from 3D Control Systems, Inc.
#
# If you do not have the license to use this software, please delete all software files
# immediately and contact sales to obtain the license: sales@3dprinteros.com.
# If you are unsure about the licensing please contact directly our sales: sales@3dprinteros.com.

import os
import sys
import logging
import subprocess
import signal

import config
import platforms

def remove_non_existed_modules(camera_modules_dict):
    working_dir = os.path.dirname(os.path.abspath(__file__))
    verified_camera_modules = {}
    for camera_name,module in camera_modules_dict.iteritems():
        if not module or os.path.exists(os.path.join(working_dir, module)):
            verified_camera_modules[camera_name] = module
    return verified_camera_modules

class CameraController:

    HD_CAMERA_NAME = 'HD camera'
    CAMERA_MODULES = { "Dual camera": "dual_cam.py",
                       "Multi camera": "multi_cam.py",
                       "Pi camera": "rpi_cam.py",
                       "Disable camera": None }
    if platforms.get_platform() == 'win':
        CAMERA_MODULES[HD_CAMERA_NAME] = "hdcamera.exe"
    else:
        CAMERA_MODULES[HD_CAMERA_NAME] = "hdcamera"
    CAMERA_MODULES = remove_non_existed_modules(CAMERA_MODULES)
    HD_CAMERA_URL = 'hdcam-' + config.get_settings()['URL']
    HD_CAMERA_PORT = '443'

    def __init__(self, token, mac):
        self.logger = logging.getLogger(__name__)
        self.token = token
        self.mac = mac
        self.camera_process = None
        self.current_camera_name = "Disable camera"
        self.terminating_count = 0
        self.start_camera_process()

    def start_camera_process(self, camera_name=None, token=None):
        self.logger.info('Launching camera subprocess')
        settings = config.get_settings()
        camera_name_default = settings['camera']['default']
        if not camera_name:
            camera_name = camera_name_default
        module_name = self.CAMERA_MODULES[camera_name]
        if camera_name_default != camera_name:
            settings['camera']['default'] = camera_name
            config.Config.instance().save_settings(settings)
        if not settings["camera"]["enabled"]:
            self.logger.info("Can't launch camera - disabled in config")
            return False
        if not token:
            token = self.token
        if module_name:
            cam_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), module_name)
            if cam_path.endswith('.py'):
                start_command = [sys.executable, cam_path, token, self.mac]
            else:
                start_command = [cam_path, self.HD_CAMERA_URL, self.HD_CAMERA_PORT, token]
            try:
                self.camera_process = subprocess.Popen(start_command, close_fds=True)
            except Exception as e:
                self.logger.warning('Could not launch camera due to error:\n' + str(e))
            else:
                self.current_camera_name = camera_name
                self.token = token
                self.logger.info('Camera started: ' + camera_name)
                return True

    def restart_camera(self):
        self.logger.info('Restarting camera module: ' + self.current_camera_name)
        current_camera_name = self.current_camera_name
        if self.stop_camera_process():
            return self.start_camera_process(current_camera_name)

    def get_current_camera_name(self):
        return self.current_camera_name

    def switch_camera(self, new_camera_name, token):
        if new_camera_name != self.HD_CAMERA_NAME: #for starting camera by command from base sender
            token = config.get_app().user_login.user_token
        self.logger.info('Switching camera module from %s to %s' % (self.current_camera_name, new_camera_name))
        if self.stop_camera_process():
            return self.start_camera_process(camera_name=new_camera_name, token=token)

    def stop_camera_process(self):
        self.terminating_count += 1
        terminating_count = self.terminating_count
        if self.camera_process:
            self.logger.info('Terminating camera process...')
            #if platforms.PLATFORM == 'win':
                #self.camera_process.send_signal(signal.CTRL_C_EVENT)
            #else:
            self.camera_process.terminate()
            while True:
                try:
                    if self.camera_process.poll() != None:
                        break
                except (OSError, AttributeError):
                    break
            self.logger.info('...camera process terminated.')
            self.current_camera_name = "Disable camera"
            self.camera_process = None
        if self.terminating_count == terminating_count:
            self.terminating_count = 0
            return True
