#
# Copyright 3D Control Systems, Inc. All Rights Reserved 2017-2019. Built in San Francisco.
#
# This software is distributed under commercial non-GPL license for personal, educational,
# corporate or any other use. The software as a whole or any parts of that are prohibited
# for distribution and/or use without obtaining license from 3D Control Systems, Inc.
#
# If you do not have the license to use this software, please delete all software files
# immediately and contact sales to obtain the license: sales@3dprinteros.com.
# If you are unsure about the licensing please contact directly our sales: sales@3dprinteros.com.

import os
import json
import threading

import paths

REMOTE_IP = ""
LOCAL_IP = "127.0.0.1"
USER_SETTINGS_FLUSH = "first_run_flush"


def get_settings():
    return Config.instance().settings


def get_profiles():
    return Config.instance().profiles


def get_app():
    return Config.instance().app


def get_error_reports():
    return [pi.errors for pi in get_app().printer_interfaces]


def merge_dictionaries(base_dict, update_dict, overwrite = False):
    result_dict = {}
    result_dict.update(base_dict)
    for key, value in update_dict.items():
        value_to_update = result_dict.get(key)
        if type(value_to_update) != type(value):
            result_dict[key] = value
        elif isinstance(value_to_update, dict) and isinstance(value, dict):
            result_dict[key] = merge_dictionaries(value_to_update, value)
        elif overwrite and value_to_update != value:
            result_dict[key] = value_to_update
    return result_dict


class Singleton(object):
    lock = threading.Lock()
    _instance = None

    @classmethod
    def instance(cls):
        with cls.lock:
            if not cls._instance:
                print "Creating new instance of " + cls.__name__
                cls._instance = cls()
        return cls._instance


class Config(Singleton):

    DEFAULT_SETTINGS_FILE = os.path.join(os.path.dirname(__file__), 'default_settings.json')
    USER_SETTINGS_FILE = os.path.join(paths.current_settings_folder(), 'user_settings.json')

    def __init__(self):
        self.init_settings()
        self.profiles = []
        self.app = None

    def init_settings(self):
        initial_settings = self.load_settings(self.USER_SETTINGS_FILE, warnings = False)
        settings = self.load_settings(self.DEFAULT_SETTINGS_FILE)
        if initial_settings:
            if not self.check_and_flush(initial_settings, settings):
                settings = merge_dictionaries(initial_settings, settings)
        if settings != initial_settings:
            self.save_settings(settings, self.USER_SETTINGS_FILE)
        self.settings = settings

    def load_settings(self, settings_path, warnings=True):
        try:
            with open(settings_path) as settings_file:
                config = json.loads(settings_file.read())
        except Exception as e:
            if warnings:
                print "Error reading config from %s: %s" % (settings_path, str(e))
        else:
            return config

    def save_settings(self, settings, settings_path=None):
        if not settings_path:
            settings_path=self.USER_SETTINGS_FILE
        try:
            with open(settings_path, 'w') as settings_file:
                json_config = json.dumps(settings, sort_keys = True, indent = 4, separators = (',', ': '))
                settings_file.write(json_config)
        except Exception as e:
            print "Error writing config to %s: %s" % (settings_path, str(e))
        else:
            self.settings = settings

    def restore_default_settings(self):
        if os.path.exists(self.USER_SETTINGS_FILE):
            try:
                os.remove(self.USER_SETTINGS_FILE)
            except Exception as e:
                return str(e)
        self.settings = self.load_settings(self.DEFAULT_SETTINGS_FILE)

    def set_profiles(self, profiles):
        self.profiles = profiles

    def set_app_pointer(self, app):
        self.app = app
    
    def check_and_flush(self, initial_settings, settings):
        if settings.get(USER_SETTINGS_FLUSH) and not initial_settings.get(USER_SETTINGS_FLUSH):
            print "Settings folder flush initiated by %s field" % USER_SETTINGS_FLUSH
            for filename in os.listdir(paths.CURRENT_SETTINGS_FOLDER):
                filepath = os.path.join(paths.CURRENT_SETTINGS_FOLDER, filename)
                try:
                    if os.path.isfile(filepath):
                        os.remove(filepath)
                except:
                    print "Failed to remove file: " + filepath
            return True
