#
# Copyright 3D Control Systems, Inc. All Rights Reserved 2017-2019. Built in San Francisco.
#
# This software is distributed under commercial non-GPL license for personal, educational,
# corporate or any other use. The software as a whole or any parts of that are prohibited
# for distribution and/or use without obtaining license from 3D Control Systems, Inc.
#
# If you do not have the license to use this software, please delete all software files
# immediately and contact sales to obtain the license: sales@3dprinteros.com.
# If you are unsure about the licensing please contact directly our sales: sales@3dprinteros.com.

import os
import logging
import urllib
import time
import json
import zipfile
import httplib
import threading

import version
import config
import log
import platforms
import paths


class Updater:

    IDLE_TIMEOUT_FOR_AUTOUPDATE = 120

    def __init__(self, parent):
        self.logger = logging.getLogger(__name__)
        self.update_download_url = None
        self.update_available = False
        self.check_time = 0
        self.updating_thread = None
        self.parent = parent
        self.URL = "%s-update.%s.%s" % tuple(config.get_settings()['URL'].split('.'))

    def check_update_timer(self):
        current_time = time.time()
        if current_time - self.check_time > config.get_settings()['update']['check_pause']:
            self.check_time = current_time
            self.check_for_updates()

    def check_for_updates(self):
        if config.get_settings()['update']['enabled']:
            self.update_download_url = self.request_update_download_url()
            if self.update_download_url:
                self.logger.info('Updates available!')
                self.update_available = True
                if config.get_settings()['update']['auto_update_enabled']:
                    if not (self.updating_thread and self.updating_thread.is_alive()):
                        self.updating_thread = threading.Thread(target = self.auto_update)
                        self.updating_thread.start()

    def request_update_download_url(self):
        payload = json.dumps( { 'platform': platforms.get_platform(), 'client_version': version.version,
                                'branch': version.branch, 'build': version.build } )
        headers = {"Content-Type": "application/json", "Content-Length": str(len(payload))}
        try:

            connection = httplib.HTTPConnection(self.URL, timeout=2)
            connection.connect()
            connection.request('POST', '/', payload, headers)
            resp = connection.getresponse()
            update_download_url = resp.read()
            connection.close()
        except Exception as e:
            self.logger.warning('Unable to connect to updater server. ' + str(e))
        else:
            if update_download_url:
                print update_download_url
                self.logger.debug('Update address received: ' + update_download_url)
                return update_download_url

    @log.log_exception
    def auto_update(self):
        self.logger.info('Starting automatic update')
        error = self.download_update()
        if not error:
            self.wait_for_prints_end()

    def wait_for_prints_end(self):
        self.logger.info('Waiting for prints to end to install update...')
        last_non_idle_time = time.time()
        while not self.parent.stop_flag:
            for printer_interface in config.get_app().printer_interfaces:
                if printer_interface.get_printer_state() != 'ready':
                    last_non_idle_time = time.time()
            if time.time() - last_non_idle_time > self.IDLE_TIMEOUT_FOR_AUTOUPDATE:
                self.logger.info('...all prints are finished. Rebooting client...')
                config.get_app().stop_flag = True
            time.sleep(3)

    def download_update(self):
        if self.update_available:
            if os.path.exists(paths.UPDATE_FILE_PATH) and zipfile.is_zipfile(paths.UPDATE_FILE_PATH):
                self.logger.info('Update package already downloaded')
            else:
                self.logger.info('Downloading update package...')
                try:
                    urllib.urlretrieve(self.update_download_url, paths.UPDATE_FILE_PATH)
                except Exception as e:
                    error_message = 'Update download failed: error while downloading from %s\nDetails: %s' %\
                                        ( self.update_download_url, str(e) )
                    self.logger.warning(error_message)
                    return error_message
                else:
                    if zipfile.is_zipfile(paths.UPDATE_FILE_PATH):
                        self.logger.info('...update successfully downloaded!')
                        self.update_available = False
                        # log.send_logs()
                    else:
                        error_message = 'Error: corrupted update package.'
                        self.logger.warning(error_message)
                        try:
                            os.remove(paths.UPDATE_FILE_PATH)
                        except:
                            self.logger.warning("Unable to remove corrupted update package.")
                        else:
                            self.logger.warning("Corrupted update package removed.")
                        return error_message


if __name__ == "__main__":

    class FakeApp:
        def __init__(self):
            self.stop_flag = False

    fa = FakeApp()
    u = Updater(fa)
    print u.request_update_download_url()