#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 3D Control Systems, Inc. All Rights Reserved 2017-2019. Built in San Francisco.
#
# This software is distributed under commercial non-GPL license for personal, educational,
# corporate or any other use. The software as a whole or any parts of that are prohibited
# for distribution and/or use without obtaining license from 3D Control Systems, Inc.
#
# If you do not have the license to use this software, please delete all software files
# immediately and contact sales to obtain the license: sales@3dprinteros.com.
# If you are unsure about the licensing please contact directly our sales: sales@3dprinteros.com.

import re
import string
import logging
import time

import usb.core
import usb.util
import usb.backend.libusb1
import serial.tools.list_ports
import paths
import config


class USBDetector:

    vid_pid_re = re.compile(
        '(?:.*\=([0-9-A-Z-a-f]+):([0-9-A-Z-a-f]+))|(?:.*VID_([0-9-A-Z-a-f]+)\+PID_([0-9-A-Z-a-f]+)\+)')
    serial_number_re = re.compile('(?:.*SNR\=([0-9-A-Z-a-f]+).*)|(?:.*PID_[0-9-A-Z-a-f]+\+([0-9-A-Z-a-f]+).*)')

    @staticmethod
    def format_vid_or_pid(vid_or_pid):
        return hex(vid_or_pid)[2:].zfill(4).upper()

    def detect_devices(self):
        retries = 5
        while retries: # the retries is for mac, because it can suddenly refuse to detect usb devices
            try:
                backend_from_our_directory = usb.backend.libusb1.get_backend(find_library=paths.get_libusb_path)
                devices = usb.core.find(find_all=True, backend=backend_from_our_directory)
            except Exception:
                logging.getLogger(self.__class__.__name__).exception("Exception in pyusb: ")
                devices = None
            if devices:
                return devices
            message = "Warning: no USB devices detected. This is an error. Detection retries left: %d" % retries
            logging.getLogger(self.__class__.__name__).warning(message)
            time.sleep(0.5)
            retries -= 1
        else:
            return []

    def get_printers_list(self):
        try:
            serial_ports = list(serial.tools.list_ports.comports())
        except Exception as e:
            logging.getLogger(self.__class__.__name__).warning(e)
            return
        unused_serial_ports = filter(lambda x: x[2] != "n/a", serial_ports)
        printers = []
        for dev in self.detect_devices():
            VID = USBDetector.format_vid_or_pid(dev.idVendor) #cuts "0x", fill with zeroes if needed, doing case up
            PID = USBDetector.format_vid_or_pid(dev.idProduct)
            profile = self.find_profile_by_vid_pid(VID, PID)
            if profile: # Warning! This is can get a profile of another(similar) printer type. Use with caution.
                if profile.get('crazy_serial_number'):
                    SNR = None
                else:
                    try:
                        SNR = usb.util.get_string(dev, dev.iSerialNumber)
                        for symbol in SNR:
                            if not symbol in string.printable:
                                raise RuntimeError
                    except:
                        SNR = None
                usb_info = {'VID': VID, 'PID': PID, 'SNR': SNR}
                if SNR:
                    usb_info['COM'] = self.get_serial_port_name(unused_serial_ports, VID, PID, SNR)
                printers.append(usb_info)
        for printer in printers:
            if not printer.get('COM'):
                printer['COM'] = self.get_serial_port_name(unused_serial_ports, printer['VID'], printer['PID'])
            if not printer['SNR']:
                printer['SNR'] = self.get_snr_by_serial_port_name(printer['COM'], serial_ports)
        self.modify_identical_printers_id(printers)
        return printers

    def get_snr_by_serial_port_name(self, serial_port_name, all_serial_ports):
        for port_dct in all_serial_ports:            
            if port_dct[0] == serial_port_name:
                vid_pid_snr_string = port_dct[2]
                match = self.serial_number_re.match(vid_pid_snr_string)
                if match:
                    for group in match.groups():
                        if group:
                            return group

    def get_serial_port_name(self, unused_serial_ports, vid, pid, snr=None):
        for port_dct in unused_serial_ports:
            match = self.vid_pid_re.match(port_dct[2])
            if match:
                vid_of_comport = match.group(1)
                pid_of_comport = match.group(2)
                if not vid_of_comport or not pid_of_comport:
                    vid_of_comport = match.group(3)
                    pid_of_comport = match.group(4)
                vid_of_comport = vid_of_comport.zfill(4).upper()
                pid_of_comport = pid_of_comport.zfill(4).upper()
                if vid == vid_of_comport and pid == pid_of_comport:
                    if snr and not 'SNR=' + snr.upper() in port_dct[2].upper():
                        continue
                    unused_serial_ports.remove(port_dct)
                    return port_dct[0]

    def find_profile_by_vid_pid(self, vid, pid):
        profiles = config.get_profiles()
        if profiles:
            for profile in profiles:
                if [vid, pid] in profile[u"vids_pids"]:
                    return profile
        else:
            return { u"vids_pids": [[vid, pid]] } # fake profile for testing without getting profiles from server

    # find printers with same VID, PID, SNR and include COM name in their SNRs
    def modify_identical_printers_id(self, printers):
        for checked_printer in printers:
            for printer in printers:
                if printer['VID'] == checked_printer['VID'] and printer['PID'] == checked_printer['PID'] and\
                   printer['SNR'] == checked_printer['SNR'] and printer['COM'] != checked_printer['COM']:
                      self.update_duplicate_printer_info(printer)
                      self.update_duplicate_printer_info(checked_printer)

    def update_duplicate_printer_info(self, printer):
        printer['ID_CLONE_WARNING'] = True
        if printer['COM']:
            if not printer['SNR']:
                printer['SNR'] = ''
            printer['SNR'] = printer['SNR'] + printer['COM'].split("/")[-1]


if __name__ == '__main__':
    detector = USBDetector()
    printers = detector.get_printers_list()
    print "\nAll devices:"
    for printer in printers:
        print printer
    printers = filter(lambda x: x.get('COM'), printers)
    print "\nDevices with serial port:"
    for printer in printers:
        print printer
