#
# Copyright 3D Control Systems, Inc. All Rights Reserved 2017-2019. Built in San Francisco.
#
# This software is distributed under commercial non-GPL license for personal, educational,
# corporate or any other use. The software as a whole or any parts of that are prohibited
# for distribution and/or use without obtaining license from 3D Control Systems, Inc.
#
# If you do not have the license to use this software, please delete all software files
# immediately and contact sales to obtain the license: sales@3dprinteros.com.
# If you are unsure about the licensing please contact directly our sales: sales@3dprinteros.com.

import os
import ssl
import logging
import threading
import BaseHTTPServer
from SocketServer import ThreadingMixIn
from threading import Thread

import config
import page_former
import web_interface_handler
import log



class ThreadedHTTPServer(ThreadingMixIn, BaseHTTPServer.HTTPServer):
    """ This class allows to handle requests in separated threads.
        No further content needed, don't touch this. """


class WebInterface(threading.Thread):

    @staticmethod
    def get_web_content_files(path=None):
        binary_files_extensions = ('jpg', 'png', 'ico')
        if not path:
            path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'web_content_files')
        web_content_files = {}
        for name in os.listdir(path):
            file_path = os.path.join(path, name)
            if os.path.isdir(file_path):
                web_content_files[name] = WebInterface.get_web_content_files(file_path)
            else:
                read_mode = 'r'
                if name.split(".")[-1] in binary_files_extensions:
                    read_mode += 'b'
                try:
                    with open(file_path, read_mode) as f:
                        web_content_files[name] = f.read()
                except Exception as e:
                    logger = logging.getLogger(__name__)
                    error = 'Error while reading file "' + name + '": ' + str(e)
                    logger.error(error)
                    raise RuntimeError(error)
        return web_content_files

    def __init__(self, app):
        self.logger = logging.getLogger(__name__)
        self.app = app
        self.server = None
        self.server_https = None
        self.web_content_files = {}
        self.thread_https = None
        threading.Thread.__init__(self, name="WebInterface")

    def find_printer_interface(self, usb_info_string):
        for printer_interface in self.server.app.printer_interfaces:
            if str(printer_interface.usb_info) == usb_info_string:
                return printer_interface

    @log.log_exception
    def run(self):
        self.logger.info("Starting web server...")
        if config.get_settings()['remote_control']['web_server']:
            IP = config.REMOTE_IP
        else:
            IP = config.LOCAL_IP
        port = config.get_settings()['web_interface']['port']
        port_https = config.get_settings()['web_interface']['port_https']
        # self.web_content_files = self.get_web_content_files()
        if port_https:
            self.thread_https = Thread(target=self.serve_on_port, args=[port_https, IP, 'server_https', True])
            self.thread_https.start()
        self.serve_on_port(port, IP)
        if self.thread_https:
            try:
                self.thread_https.join()
            except AttributeError:
                pass

    @log.log_exception
    def serve_on_port(self, port, IP, class_property='server', https=False):
        try:
            server = ThreadedHTTPServer((IP, port), web_interface_handler.WebInterfaceHandler)
            server.is_https = https
            server.daemon_threads = True
            if https and not self.use_ssl_certificates(server):
                return
            setattr(self, class_property, server)
        except Exception as e:
            self.logger.error(e)
        else:
            self.logger.info("...web server started on port: %s%s" % (port, " (HTTPS)" if https else ""))
            server.app = self.app
            server.web_content_files = self.get_web_content_files()
            server.page_former = page_former.PageFormer(server)
            server.find_printer_interface = self.find_printer_interface
            server.serve_forever()
            if server.socket is not None:
                server.server_close()
            self.logger.info("Web server stopped on port: %s" % port)

    def use_ssl_certificates(self, server):
        import paths
        import os.path
        keyfile = os.path.join(paths.current_settings_folder(), 'private_key.pem')
        certfile = os.path.join(paths.current_settings_folder(), 'certificate.pem')
        if not os.path.isfile(keyfile) or not os.path.isfile(certfile):
            cmd = ["openssl", "req", "-x509", "-newkey", "rsa:2048", "-keyout", keyfile, "-out", certfile,
                   "-days", "10000", "-nodes", "-subj",
                   "/C=US/ST=California/L=San Francisco/O=3D Control Systems/OU=3DPrinterOS/CN=3dprinteros.com"]
            try:
                import subprocess
                call = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                stdout, stderr = call.communicate()
                self.logger.info('openssl stdout: %s' % stdout)
                self.logger.info('openssl stderr: %s' % stderr)
                if call.returncode != 0:
                    self.logger.error('Some problem with auto certificates generating. '
                                      'Please refer to how-to-generate-certificates.txt and generate certs manually.')
                    return False
            except Exception as e:
                self.logger.error('openssl certificates generating error: %s' % e)
                return False
        ssl_version = ssl.PROTOCOL_SSLv23
        if hasattr(ssl, 'PROTOCOL_TLSv1_2'):
            ssl_version = ssl.PROTOCOL_TLSv1_2
        server.socket = ssl.wrap_socket(server.socket, ssl_version=ssl_version, keyfile=keyfile,
                                        certfile=certfile, ca_certs=None, server_side=True,
                                        ciphers='HIGH:!aNULL:!eNULL:!SSLv2:!SSLv3:!MD5:!RC4:!DSS:!3DES:@STRENGTH')
        return True
